%%%-------------------------------------------------------------------
%%% File    : dialyzer_analysis.erl
%%% Author  : Tobias Lindahl <tobiasl@csd.uu.se>
%%% Description : Front end for the analysis employed by the dialyzer.
%%%
%%% Created : 7 May 2004 by Tobias Lindahl <tobiasl@csd.uu.se>
%%%-------------------------------------------------------------------

-module(dialyzer_analysis).
-export([start/7]).

-define(OPTS(Pid_), 
	[{type_warnings, {pid, Pid_}}, type_only, no_inline_fp,
	 {use_callgraph, fixpoint}, get_called_modules]).


start(Files, Dets, Parent, MCG, Fixpoint1, InitDets, FileType) ->
  Opts1 = 
    case InitDets of
      keep ->
	[{icode_type, {dets, Dets}}|?OPTS(self())];
      _ ->
	?OPTS(self())
    end,

  {Opts2, Fixpoint2} =
    case {FileType, Fixpoint1} of
      {module, false} ->
	{[use_callgraph | Opts1], false};
      {module, first} ->
	{Opts1, true};
      {application, false} ->
	{[use_callgraph | Opts1], false};
      {application, first} ->
	{[no_type_warnings|Opts1], true};
      {application, second} ->
	{Opts1, true}
    end,

  case FileType of
    module ->
      analyze_all_mods(Files, Opts2, Parent, Dets, Fixpoint2, InitDets);
    application ->
      analyze_all_apps(Files, Opts2, Parent, Dets, InitDets, MCG)
  end,
  done(Parent).


%%============ Analyze all modules ==============

analyze_all_mods(Mods, Opts, Parent, Dets, Fixpoint, InitDets) ->
  ModMap = get_modules_by_dir(Mods),
  case Fixpoint of
    true ->
      analyze_all_dirs(ModMap, [use_callgraph|Opts], Parent, Dets, InitDets);
    false ->
      analyze_all_dirs(ModMap, Opts, Parent, Dets, InitDets)
  end.

analyze_all_dirs([{Dir, Mods}|Left], Opts, Parent, Dets, InitDets) ->
  case file:set_cwd(Dir) of
    ok ->
      [analyze_mod(X, Parent, Opts, Dets, InitDets) || X <- Mods];
    {error, _Reason} ->
      Msg = io_lib:format("*****Could not cd into ~s\n", [Dir]),
      send_log(Parent, Msg)
  end,
  analyze_all_dirs(Left, Opts, Parent, Dets, InitDets);
analyze_all_dirs([], _Opts, _Parent, _Dets, _InitDets) ->
  ok.

%%============ Analyze all applications ==============

analyze_all_apps(Apps, Opts, Parent, Dets, InitDets, MCG) ->
  ModMap = get_modules_by_app(Apps),
  case analyze_all_apps(Apps, Opts, Parent, Dets, InitDets, ModMap, MCG, []) of
    ok ->
      ok;
    {module_callgraph, NewMCG} ->
      send_module_callgraph(Parent, NewMCG)
  end,
  ok.

analyze_all_apps([App|Left], Opts, Parent, Dets, InitDets, ModMap, MCG, Acc) ->
  {NewDets, NewOpts} = 
    case InitDets of
      application ->
	TmpDets = mk_tmp_dets(Dets),
	{TmpDets, [{icode_type, {dets, TmpDets}}|Opts]};
      _ ->
	{Dets, Opts}
    end,
  case analyze_app(App, ModMap, NewOpts, Parent, NewDets, InitDets, MCG) of
    error ->
      analyze_all_apps(Left, Opts, Parent, Dets, InitDets, ModMap, MCG, Acc);
    ok ->
      send_log(Parent, io_lib:format("==== Leaving ~s\n",[App])),
      analyze_all_apps(Left, Opts, Parent, Dets, InitDets, ModMap, MCG, Acc);
    {callgraph, CG} ->
      send_log(Parent, io_lib:format("==== Leaving ~s\n",[App])),
      analyze_all_apps(Left, Opts, Parent, Dets, InitDets, ModMap, MCG,
		       [{App, CG}|Acc])
  end;
analyze_all_apps([], _Opts, _Parent, _Dets, _InitDets, _ModMap, _MCG, Acc) ->
  case Acc of
    [] ->
      ok;
    List when is_list(List) ->
      {module_callgraph, orddict:from_list(List)}
  end.


%%============ Analyze one application ==============

analyze_app(App, ModMap, Opts, Parent, Dets, InitDets, UseMCG) ->
  case UseMCG of
    none ->
      analyze_app_no_callgraph(App, ModMap, Opts, Parent, Dets, InitDets);
    {module_callgraph, MCG} ->
      case orddict:find(App, MCG) of
	{ok, CG} ->
	  analyze_app_with_callgraph(App, [no_get_called_modules|Opts], 
				     Parent, Dets, CG);
	error ->
	  analyze_app_no_callgraph(App, ModMap, Opts, Parent, Dets, InitDets)
      end
  end.

analyze_app_no_callgraph(App, ModMap, Opts, Parent, Dets, InitDets) ->
  case orddict:fetch(App, ModMap) of
    [] ->
      Msg = io_lib:format("*****Skipping ~s. Contained no beam-files\n",[App]),
      send_log(Parent, Msg),
      error;
    Mods ->
      Dir = filename:join(App, "ebin"),
      case file:set_cwd(Dir) of
	ok ->
	  send_current_app(Parent, App),
	  Res = [analyze_mod(X, Parent, Opts, Dets, InitDets) || X <- Mods],
	  case lists:any(fun(X)->X==error orelse X==ok end, Res) of
	    true ->
	      ok;
	    false ->
	      {callgraph, dialyzer_module_callgraph:construct(Res)}
	  end;
	{error, _Reason} ->
	  Msg = io_lib:format("*****Skipping ~s. Could not cd into ~s\n",
			      [App, Dir]),
	  send_log(Parent, Msg),
	  error
      end
  end.

analyze_app_with_callgraph(App, Opts, Parent, Dets, Callgraph) ->
  Dir = filename:join(App, "ebin"),
  case file:set_cwd(Dir) of
    ok ->
      send_current_app(Parent, App),
      analyze_sccs(Parent, Opts, Callgraph, Dets),
      ok;
    {error, _Reason} ->
      Msg = 
	io_lib:format("*****Skipping ~s. Could not cd into ~s\n",[App, Dir]),
      send_log(Parent, Msg),
      error
  end.

%%============ Analyze Strongly Conected Components ==============

analyze_sccs(Parent, Opts, Callgraph, Dets) ->
  case dialyzer_module_callgraph:is_empty(Callgraph) of
    true ->
      ok;
    false ->
      case dialyzer_module_callgraph:take_first(Callgraph) of
	{[Mod], NewCallgraph} ->
	  %% Only one module
	  analyze_mod(Mod, Parent, Opts),
	  analyze_sccs(Parent, Opts, NewCallgraph, Dets);
	{Scc, NewCallgraph} ->
	  send_log(Parent, 
		   "-- Analyzing strongly connected component\n"
		   "-- (holding warnings)\n"),
	  analyze_one_scc(Scc, [no_type_warnings|Opts], Parent, Dets),
	  send_log(Parent,
		   "-- Reached fixpoint for strongly connected component\n"
		   "-- Re-analyzing to get warnings\n"),
	  [analyze_mod(X, Parent, Opts) || X <- Scc],
	  analyze_sccs(Parent, Opts, NewCallgraph, Dets)
      end
  end.

analyze_one_scc(Scc, Opts, Parent, Dets) ->
  case analyze_one_scc(Scc, Opts, Dets, Parent, true) of
    true -> 
      ok;
    false -> 
      send_log(Parent, "-- Re-analyzing strongly connected component --\n"),
      analyze_one_scc(Scc, Opts, Parent, Dets)
  end.

analyze_one_scc([Mod|Left], Opts, Dets, Parent, Fixpoint) ->
  Before = get_module_info(Dets, Mod),
  analyze_mod(Mod, Parent, [no_type_warnings|Opts]),
  After = get_module_info(Dets, Mod),
  case info_is_equal(Before, After) of
    true -> analyze_one_scc(Left, Opts, Dets, Parent, Fixpoint);
    false -> analyze_one_scc(Left, Opts, Dets, Parent, false)
  end;
analyze_one_scc([], _Opts, _Dets, _Parent, Fixpoint) ->
  Fixpoint.

get_module_info(Dets, Mod) ->
  {ok, _} = dets:open_file(Dets, []),
  Info = dets:match_object(Dets, {{Mod, '_', '_'}, '_'}),
  OrdDict = orddict:from_list(Info),
  dets:close(Dets),
  OrdDict.

info_is_equal([{MFA, Type1}|Left1], [{MFA, Type2}|Left2]) ->
  case erl_types:is_equal(Type1, Type2) of
    true ->
      info_is_equal(Left1, Left2);
    false ->
      false
  end;
info_is_equal([], []) ->
  true;
info_is_equal(_, _) ->
  false.

%%============ Analyze one module ==============

analyze_mod(Mod, Parent, Opts, Dets, InitDets) ->
  case InitDets of
    module ->
      TmpDets = mk_tmp_dets(Dets),
      analyze_mod(Mod, Parent, [{icode_type, {dets, TmpDets}}|Opts]);
    _ ->
      analyze_mod(Mod, Parent, Opts)
  end.

analyze_mod(Mod, Parent, Opts) ->
  send_log(Parent, "Analyzing " ++ Mod),
  case analyze_mod(Mod, Opts) of
    {error, ErrorReport} ->
      send_warning(Parent,
		   io_lib:format("Analysis failed. Please recompile ~s "
				 "using a later version "
				 "of the BEAM compiler\n", [Mod])),
      send_log(Parent, ErrorReport),
      error;
    {ok, Msg} ->
      send_all_warnings(Parent),
      send_log(Parent, Msg),
      ok;
    {ok, Msg, Callgraph} ->
      send_all_warnings(Parent),
      send_log(Parent, Msg),
      {Mod, Callgraph}
  end.


analyze_mod(Mod, Opts) ->
  T0 = time_now(),
  case catch hipe:c(list_to_atom(Mod), Opts) of
    {{ok, _}, C} ->
      T = time_since(T0) / 1000,
      {ok, io_lib:format("...done in ~.2f secs\n",[T]), C};
    {ok, _} ->
      T = time_since(T0) / 1000,
      {ok, io_lib:format("...done in ~.2f secs\n",[T])};
    {error, What}  ->
      {error, 
       io_lib:format("\n*****Analysis failed with error report:\n~p\n",[What])};
    {'EXIT', Why} ->
      {error, 
       io_lib:format("\n*****Analysis failed with error report:\n~p\n",[Why])}
  end.

time_now() ->
  T1 = hipe_bifs:get_hrvtime(),
  {time_now,T1}.

time_since({time_now,T1}) ->
  T2 = hipe_bifs:get_hrvtime(),
  T = T2-T1,
  trunc(T).

%%============ Get module names ==============

get_modules_by_app(Apps)->
  get_modules_by_app(Apps, []).

get_modules_by_app([App|Left], Acc)->
  Dir = filename:join(App, "ebin"),
  AppFile = filename:join(Dir, filename:basename(App) ++ ".app"),
  Mods =
    case file:consult(AppFile) of
      {ok,[Tuple]} ->
	{value,{modules,M}} = lists:keysearch(modules,1,element(3,Tuple)),
		[filename:basename(X) || X <- M];
      {error,_} ->
	BeamFiles =
	  case file:list_dir(Dir) of
	    {ok, F} -> 
	      lists:filter(fun(X)->filename:extension(X)=:=".beam"end,F);
	    {error, _Reason} -> 
	      []
	  end,
	[filename:basename(X, ".beam") || X <- BeamFiles]
    end,
  get_modules_by_app(Left, orddict:store(App, Mods, Acc));
get_modules_by_app([], Acc) ->
  Acc.

get_modules_by_dir(Files)->
  get_modules_by_dir(Files, []).

get_modules_by_dir([File|Left], Acc)->
  BaseName = filename:basename(File, ".beam"),
  Dir = filename:dirname(File),
  case orddict:is_key(Dir, Acc) of
    true ->
      get_modules_by_dir(Left, orddict:append(Dir, BaseName, Acc));
    false ->
      get_modules_by_dir(Left, orddict:store(Dir, [BaseName], Acc))
  end;
get_modules_by_dir([], Acc) ->
  Acc.
  

%%============ Make a temporary dets ==============

mk_tmp_dets(Dets) ->
  TmpDets = Dets++"Tmp",
  case file:copy(Dets, TmpDets) of
    {error, Reason} ->
      Msg = io_lib:format("*****Aborting! Could not make a temp copy of"
			  " the lookup table:\n ~p\n", [Reason]),
      exit({error, Msg});
    {ok, _} ->
      TmpDets
  end.


%%============ Send functions ==============

send_log(Parent, Msg) ->
  Parent ! {self(), log, Msg}.

send_warning(Parent, Msg) ->
  Parent ! {self(), warnings, Msg}.


send_all_warnings(Parent) ->
  case get_all_warnings() of
    [] -> ok;
    Warnings -> Parent ! {self(), warnings, Warnings}
  end.

get_all_warnings() ->
  receive
    {type_warning, TW} ->
      [TW | get_all_warnings()]
  after 0 ->
      []
  end.

send_current_app(Parent, App) ->
  Parent ! {self(), current_app, App}.

send_module_callgraph(Parent, MCG) ->
  Parent ! {self(), module_callgraph, MCG}.

done(Parent) ->
  Parent ! {self(), done},
  ok.
